"Test for samples module"
import numpy as np
import pandas as pd
from classifier.samples import Samples, TimeSeries
from classifier.utils.config import setup_config, SubsampleSupervisedConfiguration


def test_impute_timeseries_AsExpected(
        setup_samples_df_with_nans_ref, setup_samples_df_with_nans):
    ref_samples = setup_samples_df_with_nans_ref
    timeseries = TimeSeries(None, None, None)

    timeseries_imputed = timeseries.impute_ts(setup_samples_df_with_nans)
    assert np.array_equal(
        ref_samples.to_numpy(),
        timeseries_imputed.to_numpy())


def test_outlier_removal_RemovedAsExpected(
        setup_samples_df, setup_samples_df_with_outliers):
    config = setup_config()
    
    ref_samples = setup_samples_df
    samples_outlier_removed = Samples.outlier_removal(
        setup_samples_df_with_outliers, config)
    np.array_equal(ref_samples.values, samples_outlier_removed.values)
    
    
def test_subsample_AsExpected(
        setup_samples_df_with_rois_fid, setup_samples_df_with_rois_fid_subsampled):
    samples = setup_samples_df_with_rois_fid
    
    config = SubsampleSupervisedConfiguration(
        active=True,
        group_by='roi_fid',
        sample_type='n',
        amount=1,
        replace=False
    )

    samples_subsampled = Samples.subsample(samples, config)
    pd.testing.assert_frame_equal(samples_subsampled, setup_samples_df_with_rois_fid_subsampled)


"""File containing all model classes necessary to run the classifier"""
import logging
from pathlib import Path
from typing import Any, Callable, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.ensemble import IsolationForest, RandomForestClassifier
from sklearn.model_selection import RandomizedSearchCV
from sklearn.preprocessing import MinMaxScaler
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import Pipeline
from sklearn.base import BaseEstimator, TransformerMixin

from xgboost import XGBClassifier


# initiate logger
MODELS_LOGGER = logging.getLogger(__name__)


class BaseMixin:
    """A mixin class for all models. This allows for each algorithm to use the
    optimise and plotting functions"""

    def plot_time_vs_accuracy(self, out_dir: Path) -> None:
        """Plotting the training time vs the accuracy.

        Args:
            time (array): The time values to plot
            accuracy(array) The accuracy values to plot
            out_dir (Path): Output directory

        Returns:
            None
        """
        time = self.cv_df['mean_fit_time']
        accuracy = self.cv_df['mean_test_score']
        fig = plt.figure()
        axis = fig.add_subplot(111)
        axis.scatter(time, accuracy)
        axis.set_xlabel("Time (s)")
        axis.set_ylabel("Accuracy (-) ")
        plt.tight_layout()
        plt.savefig(out_dir / 'Optimisation_time.png', dpi=300)

    def random_optimise(
            self, optimization_parameters: dict,
            trainx: np.ndarray, trainy: np.ndarray, out_dir: Path,
            optimise_iters: int = 10) -> Any:
        """Optimization of a parameter space using random samples from the
        parameter space

        Args:
            trainx (np.ndarray):   Training dataset features (X-values)
            trainy (np.ndarray):   Training dataset outputs
                (class names, y values)
            out_dir (Path):  The output directory path
            optimise_iters (int): The number of iterations for the optimisation.

        Returns:
            The best performing parameter combination  (dict)

        """
        MODELS_LOGGER.info("\n####-----Optimization----#####\n")
        MODELS_LOGGER.info(
            "Starting Optimization. This might take a while....")
        clf = RandomizedSearchCV(
            self,
            optimization_parameters, scoring='accuracy', verbose=100,
            refit=False, cv=3, return_train_score=True, n_jobs=-1,
            n_iter=optimise_iters)

        clf.fit(trainx, trainy)
        self.cv_df = pd.DataFrame(clf.cv_results_)
        MODELS_LOGGER.debug("--Optimization Results--\n %s",
                            self.cv_df[['mean_train_score', 'mean_test_score',
                                        'mean_fit_time']])
        parameter_set = self.cv_df['params'][self.cv_df['rank_test_score'] ==
                                             1].values[0]
        MODELS_LOGGER.debug("\nThe best parameter combination is:\n %s",
                            parameter_set)
        self.plot_time_vs_accuracy(out_dir)
        return parameter_set

# pylint: disable=too-many-ancestors


class RandomForest(RandomForestClassifier, BaseMixin):
    """RandomForest class; child of RandomForestClassifier from sk-learn"""

    def __init__(self,
                 n_estimators=100,
                 criterion="gini",
                 max_depth=None,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.,
                 max_features="sqrt",
                 max_leaf_nodes=None,
                 min_impurity_decrease=0.,
                 bootstrap=True,
                 oob_score=False,
                 n_jobs=1,
                 random_state=None,
                 verbose=0,
                 class_weight=None,
                 ccp_alpha=0.0,
                 max_samples=None):
        super().__init__()

        self.n_estimators = n_estimators
        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.bootstrap = bootstrap
        self.oob_score = oob_score
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.verbose = verbose
        self.class_weight = class_weight
        self.ccp_alpha = ccp_alpha
        self.max_samples = max_samples


class XGBoost(XGBClassifier, BaseMixin):
    """XGBoost class, child of XGBoostclassifier model from sk-learn"""

    def __init__(
            self,
            objective: str ="binary:logistic",
            **kwargs: Any
        ) -> None:
        super().__init__(objective=objective, **kwargs)
        self.optimization_parameters = {
            "learning_rate": [0.1, 0.2, 0.3],
            "n_estimators": [10, 20, 50, 100, 200],
            "max_depth": [1, 3, 5, 10]
        }


class SingleClass(IsolationForest, BaseMixin):

    """Singleclass class, child of IsolationForest model from sk-learn"""

    def __init__(self,
                 n_estimators=100,
                 max_samples="auto",
                 max_features=1) -> None:
        super().__init__(
            n_estimators=n_estimators,
            max_samples=max_samples,
            max_features=max_features)

        self.optimization_parameters = {
            "n_estimators": [10, 20, 50, 100, 200],
            "max_samples": ['auto', 10, 50, 100, 200],
            "max_features": [0.1, 0.3, 0.5, 1.0]
        }


class Knn(KNeighborsClassifier, BaseMixin):
    """K-Nearest-Neighbor class; child of KNeighborsClassifier from sk-learn"""

    def __init__(self,
                 n_neighbors:int,
                 weights: Union[str, Callable],
                 algorithm: str,
                 leaf_size: int,
                 p: int,
                 metric: Union[str, Callable],
                 metric_params: dict,
                 n_jobs: int):
        super().__init__(n_neighbors=n_neighbors,
                         weights=weights,
                         algorithm=algorithm,
                         leaf_size=leaf_size,
                         p=p,
                         metric=metric,
                         metric_params=metric_params,
                         n_jobs=n_jobs)
        self.algorithm = algorithm

    def fit(self, X, y):
        if isinstance(X, tuple):
            # Tuple when patterns where created in this run from
            # DtwPatternCreator
            sample_x, sample_y = X
            super().fit(sample_x, sample_y)
        else:
            super().fit(X, y)


class KnnDtw(Pipeline, BaseMixin):
    """SKlearn Pipeline, child of Pipeline model from sk-learn
       Contains K-Nearest-Neighbors classification
       Creates time series patterns from data on the first run.
       Then normalizes the data
       And then runs classification with Knn and dtw as distance
       metrics"""

    def __init__(self,
                 n_neighbors=1,
                 weights="uniform",
                 algorithm="brute",
                 leaf_size=30,
                 p=2,
                 metric="minkowski",
                 metric_params=None,
                 n_jobs=None,
                 fit_params=None,
                 number_of_patterns_per_class=1,
                 out_dir=None,
                 patterns_path=None,
                 patterns_save=False
                 ):

        from classifier.timeseries import dtw_metric  # pylint: disable=import-outside-toplevel

        self.fit_params = fit_params
        self.n_neighbors = n_neighbors
        self.weights = weights
        self.algorithm = algorithm
        self.leaf_size = leaf_size
        self.p = p  # pylint: disable=invalid-name
        self.metric = metric
        self.metric_params = metric_params
        self.n_jobs = n_jobs
        self.number_of_patterns_per_class = \
            number_of_patterns_per_class
        self.out_dir = out_dir
        self.patterns_path = patterns_path
        self.patterns_save = patterns_save

        super().__init__(steps=[
            ('pattern_creator', DtwPatternCreator(
                number_of_patterns_per_class=number_of_patterns_per_class,
                out_dir=out_dir,
                patterns_path=patterns_path,
                patterns_save=patterns_save)),
            ('transformer', NormalizeTsTransformer(
                fit_params=fit_params)),
            ('classifier', Knn(
                n_neighbors=n_neighbors,
                weights=weights,
                algorithm=algorithm,
                leaf_size=leaf_size,
                p=p,
                metric=dtw_metric,
                metric_params=metric_params,
                n_jobs=n_jobs,
            ))
        ])

    def set_params(self, **kwargs):
        """Set the parameters of this estimator.

        Valid parameter keys can be listed with ``get_params()``.

        Returns
        -------
        self
        """
        # Add "classifier__" to Classifier arguments
        classifier_args = ["n_neighbors", "weights", "algorithm", "leaf_size",
                           "p", "metric", "metric_params", "n_jobs"]
        pattern_creator_args = ["number_of_patterns_per_class",
                                "out_dir", "patterns_path", "patterns_save"]
        transformer_args = ["fit_params"]

        classifier_kwargs = {'classifier__'+key: value for key,
                             value in kwargs.items() if key in classifier_args}
        pattern_creator_kwargs = {
            'pattern_creator__' + key: value for key, value in kwargs.items()
            if key in pattern_creator_args}
        transformer_kwargs = {
            'transformer__' + key: value for key, value in kwargs.items()
            if key in transformer_args}

        kwargs_adjusted = {
            **classifier_kwargs, **pattern_creator_kwargs, **transformer_kwargs}
        self._set_params('steps', **kwargs_adjusted)
        return self


class NormalizeTsTransformer(BaseEstimator, TransformerMixin):
    """ Normalizes timeseries data
    """
    # pylint: disable=unused-argument, invalid-name

    def __init__(self, fit_params=None):
        self.fit_params = fit_params
        self.scaler = MinMaxScaler()

    def fit(self, sample, y=None):
        """ Fit method """
        if isinstance(sample, tuple):
            # Tuple when patterns where created in this run from
            # DtwPatternCreator
            X = sample[0]
            self.scaler.fit(
                X.reshape(-1, self.fit_params['band_count']))
        else:
            self.scaler.fit(
                sample.reshape(-1, self.fit_params['band_count']))
        return self

    def transform(self, sample):  # pylint: disable=unused-argument
        """ Normalize the samples based on the limits learned on fit

        Args:
            dataset (np.ndarray):
                index: pixels; columns: bands per date (flattened)

        Returns:
            dataset_extracted_features (np.ndarray): standardized per band
        """
        if isinstance(sample, tuple):
            # Tuple when patterns where created in this run from
            # DtwPatternCreator
            sampleX, sample_y = sample
            sampleX_transformed = self.scaler.transform(
                sampleX.reshape(-1, self.fit_params['band_count'])).reshape(
                    -1, sampleX.shape[-1])
            return sampleX_transformed, sample_y

        sample_transformed = self.scaler.transform(
            sample.reshape(-1, self.fit_params['band_count'])).reshape(
                -1, sample.shape[-1])
        return sample_transformed


class DtwPatternCreator(BaseEstimator, TransformerMixin):
    """ Creates the pattern for the dynamic time warping algorithm
    """
    # pylint: disable=unused-argument, invalid-name

    def __init__(self, number_of_patterns_per_class=1,
                 out_dir=None, patterns_path=None, patterns_save=False):
        self.number_of_patterns_per_class = number_of_patterns_per_class
        self.created_patterns = False
        self.out_dir = out_dir
        self.patterns_path = patterns_path
        self.patterns_save = patterns_save

    def fit(self, sample, y=None):
        """ Fit method """
        return self

    def transform(self, sample):
        """ Transforms the data to patterns for each class. Does only transform
        once during fit.

        Args:
            dataset (np.ndarray):
                index: pixels; columns: bands per date (flattened)

        Returns:
            dataset_extracted_features (np.ndarray): standardized per band
        """
        from classifier.timeseries import get_characteristic_timeseries  # pylint: disable=import-outside-toplevel

        if not self.created_patterns:
            # Get patterns
            if self.patterns_path:
                MODELS_LOGGER.info("Use provided pattern file.")
                # From provided file
                sample = pd.read_pickle(self.patterns_path)
            else:
                MODELS_LOGGER.info("Create new patterns from data.")
                # Create new patterns
                sample = get_characteristic_timeseries(
                    sample,
                    self.number_of_patterns_per_class,
                    self.out_dir
                )
            # Save patterns only if not provided
            if self.patterns_save and not self.patterns_path and \
               self.out_dir:
                sample.to_pickle(self.out_dir / "training_patterns.pkl")

            # set y to class values of the patterns
            y = sample['class'].values
            # only extract data columns
            sample = sample.loc[:, ~sample.columns.get_level_values(0).isin(
                ['class', 'roi_fid'])].values
            self.created_patterns = True
            return sample, y
        if isinstance(sample, pd.DataFrame):
            # if it has already extracted patterns but gets dataframe with
            # class and roi_fid columns
            sample = sample.loc[:, ~sample.columns.get_level_values(0).isin(
                ['class', 'roi_fid'])].values
        # If numpy array or converted to numpy array
        return sample



"Fixture for setting up the tests, yield CliRunner() and tearing them down"
from collections import OrderedDict
import shutil
import os
from pathlib import Path
import fiona
import numpy as np
import pandas as pd
import pytest
import rasterio

from click.testing import CliRunner
from classifier.settings import WORKSPACE

from classifier.utils.general import read_config


@pytest.fixture
def runner(test_arg):
    """Setup, yield CliRunner() and teardown"""
    # setup
    is_timeseries_test = test_arg[0]
    is_supervised_test = test_arg[1]
    rois_outside_test = test_arg[2]

    # Rasters
    nowhere = 0.0
    pixel_size = 2.0

    transform = rasterio.Affine(
        pixel_size,
        nowhere,
        nowhere,
        nowhere,
        -pixel_size,
        nowhere
    )

    raster_folder = WORKSPACE / "integration_test_rasters"
    raster_folder.mkdir()

    if is_timeseries_test:
        raster_names = ['test_raster_008570_2020-07-01T000000Z_mean.tif',
                        'test_raster_008570_2020-08-01T000000Z_mean.tif',
                        'test_raster_008570_2020-09-01T000000Z_mean.tif']
    else:
        raster_names = ['test_raster_008570_2020-07-01T000000Z_mean.tif']
    raster_paths = [raster_folder / raster_name
                    for raster_name in raster_names]

    raster_values = [
        np.array([
            np.full((10, 10), [1, 1, 1, 1, 1, 100, 100, 100, 100, 100]),
            np.full((10, 10), [2, 2, 2, 2, 2, 101, 101, 101, 101, 101])
        ]),
        np.array([
            np.full((10, 10), [2, 2, 2, 2, 2, 102, 102, 102, 102, 102]),
            np.full((10, 10), [4, 4, 4, 4, 4, 103, 103, 103, 103, 103])
        ]),
        np.array([
            np.full((10, 10), [1, 1, 1, 1, 1, 103, 103, 103, 103, 103]),
            np.full((10, 10), [3, 3, 3, 3, 3, 105, 105, 105, 105, 105])
        ]),
    ]

    for raster_path, data in zip(raster_paths, raster_values):
        with rasterio.open(
            raster_path,
            'w',
            driver='GTiff',
            height=data.shape[1],
            width=data.shape[2],
            count=data.shape[0],
            dtype='float32',
            crs='+proj=latlong',
            nodata=np.nan,
            transform=transform
        ) as dst:
            dst.write(data)

    if is_supervised_test:
        # Train polygons
        x_min, y_min, x_max, y_max = 0, -20, 10, 0
        window_polygon_1 = [[(x_min, y_min),
                             (x_min, y_max),
                             (x_max, y_max),
                             (x_max, y_min)]]

        if rois_outside_test:
            # partially outside
            x_min, y_min, x_max, y_max = 10, -20, 20, 0
            window_polygon_2 = [[(x_min, y_min),
                                 (x_min, y_max),
                                 (x_max, y_max),
                                 (x_max, y_min)]]
            # completely outside
            x_min, y_min, x_max, y_max = 30, -40, 40, -20
            window_polygon_3 = [[(x_min, y_min),
                                 (x_min, y_max),
                                 (x_max, y_max),
                                 (x_max, y_min)]]

        else:
            x_min, y_min, x_max, y_max = 10, -20, 20, 0
            window_polygon_2 = [[(x_min, y_min),
                                 (x_min, y_max),
                                 (x_max, y_max),
                                 (x_max, y_min)]]

        polygon1 = {
            "properties": {
                "id": 0,
                "fid": 1
            },
            "geometry": {
                "coordinates": window_polygon_1,
                "type": "Polygon"
            }
        }
        polygon2 = {
            "type": "Feature",
                    "properties": {
                        "id": 1,
                        "fid": 2
                    },
            "geometry": {
                        "coordinates": window_polygon_2,
                        "type": "Polygon"
                    }
        }

        if rois_outside_test:
            polygon3 = {
                "type": "Feature",
                "properties": {
                        "id": 1,
                        "fid": 3
                },
                "geometry": {
                    "coordinates": window_polygon_3,
                    "type": "Polygon"
                }
            }
            train_polygons = [polygon1, polygon2, polygon3]
        else:
            train_polygons = [polygon1, polygon2]

        schema = {
            "geometry": "Polygon",
            "properties": OrderedDict([("id", "int"), ("fid", "int")])
        }
        rois_path = WORKSPACE / 'integration_test_rois.gpkg'
        with fiona.open(rois_path,
                        "w",
                        driver="GPKG",
                        crs="+proj=latlong",
                        schema=schema) as src:
            src.writerecords(train_polygons)

    yield CliRunner()

    # teardown
    shutil.rmtree(WORKSPACE / "integration_test_rasters")
    if is_supervised_test:
        os.remove(WORKSPACE / 'integration_test_rois.gpkg')
    # Delete classification result
    shutil.rmtree(WORKSPACE / "test_result")


@pytest.fixture
def setup_config():
    return read_config(Path(""))


@pytest.fixture
def setup_samples_df():
    data = np.array([[1.51, 4.57, 1],
                     [1.46, 4.61, 1],
                     [1.52, 4.53, 1],
                     [4.01, 2.03, 2],
                     [4.01, 2.01, 2],
                     [4.01, 2.02, 2],
                     ])
    columns = ("test_raster_008570_2020-07-01T000000Z_mean.tif_1",
               "test_raster_008570_2020-07-01T000000Z_mean.tif_2", "class")
    samples_test = pd.DataFrame(data, columns=columns)
    return samples_test


@pytest.fixture
def setup_samples_df_with_outliers():
    data = np.array([[1.51, 4.57, 1],
                     [1.46, 4.61, 1],
                     [6.04, 4.54, 1],  # outlier
                     [1.52, 4.53, 1],
                     [4.01, 2.03, 2],
                     [4.01, 2.01, 2],
                     [4.01, 2.02, 2],
                     [4.01, 5.02, 2],  # outlier
                     ])
    columns = ("test_raster_008570_2020-07-01T000000Z_mean.tif_1",
               "test_raster_008570_2020-07-01T000000Z_mean.tif_2", "class")
    samples_test = pd.DataFrame(data, columns=columns)
    return samples_test


@pytest.fixture
def setup_samples_df_with_nans():
    data = np.array([[1., 4., 2., np.NaN, 3., np.NaN],
                     [np.NaN, 4., 2., 6., 3., 8.],
                     [1., 4., 2., 6., 3., np.NaN],
                     [1., 4., np.NaN, 6., 3., 8.],
                     ])

    roi_classes = [0, 0, 1, 1]
    roi_fids = [1, 1, 2, 2]
    pixel_ids = [0, 1, 2, 3]
    dates = pd.to_datetime(['2020-07-01', '2020-08-01', '2020-09-01'])
    bands = ['B01', 'B02']

    index = pd.MultiIndex.from_tuples(
        list(zip(roi_classes, roi_fids, pixel_ids)),
        names=['class', 'roi_fid', 'pixel_id']
    )
    columns = pd.MultiIndex.from_product(
        [dates, bands],
        names=['Date', 'Band']
    )
    samples_test = pd.DataFrame(data, index=index, columns=columns)
    return samples_test


@pytest.fixture
def setup_samples_df_with_nans_ref():
    data = np.array([[1., 4., 2., 4., 3., 4.],
                     [2., 4., 2., 6., 3., 8.],
                     [1., 4., 2., 6., 3., 6.],
                     [1., 4., 2., 6., 3., 8.],
                     ])

    roi_classes = [0, 0, 1, 1]
    roi_fids = [1, 1, 2, 2]
    pixel_ids = [0, 1, 2, 3]
    dates = pd.to_datetime(['2020-07-01', '2020-08-01', '2020-09-01'])
    bands = ['B01', 'B02']

    index = pd.MultiIndex.from_tuples(
        list(zip(roi_classes, roi_fids, pixel_ids)),
        names=['class', 'roi_fid', 'pixel_id']
    )
    columns = pd.MultiIndex.from_product(
        [dates, bands],
        names=['Date', 'Band']
    )
    samples_test = pd.DataFrame(data, index=index, columns=columns)
    return samples_test

@pytest.fixture
def setup_samples_df_with_rois_fid():
    data = np.array([[1., 4.],
                    [2., 3.],
                    [1., 7.],
                    [1., 8.],
                    ])
        
    roi_classes = [0, 0, 1, 1]
    roi_fids = [1, 1, 2, 2]
    pixel_ids = [0, 1, 0, 1]
    bands = ['B01', 'B02']

    index = pd.MultiIndex.from_tuples(
        list(zip(roi_classes, roi_fids, pixel_ids)),
        names=['class', 'roi_fid', 'pixel_id']
    )
    columns = ['band0', 'band1']
    samples_test = pd.DataFrame(data, index=index, columns=columns)
    return samples_test

@pytest.fixture
def setup_samples_df_with_rois_fid_subsampled():
    data = np.array([[2., 3.],
                    [1., 7.],
                    ])
        
    roi_classes = [0, 1]
    roi_fids = [1, 2]
    pixel_ids = [1, 0]
    bands = ['B01', 'B02']

    index = pd.MultiIndex.from_tuples(
        list(zip(roi_classes, roi_fids, pixel_ids)),
        names=['class', 'roi_fid', 'pixel_id']
    )
    columns = ['band0', 'band1']
    samples_test = pd.DataFrame(data, index=index, columns=columns)
    return samples_test



from unittest import TestCase

import numpy as np
import pandas as pd

from classifier.utils.raster import ndarray_to_df


class NdarrayToDfTestCase(TestCase):
    def setUp(self):
        self.test_function = ndarray_to_df

    def test_convert_ndarray_to_df_ok(self):
        # arrange
        array = np.array([[[1,2], [3,4]], [[5,6], [7,8]]])
        # act
        result = self.test_function(array)
        # assert
        expected_result = pd.DataFrame([[1,5], [2,6], [3,7], [4,8]])
        self.assertTrue(result.equals(expected_result))

    def test_convert_ndarray_to_df_with_nan_ok(self):
        # arrange
        array = np.array([[[1,2], [3,4]], [[5,6], [7,np.nan]]])
        expected_result = pd.DataFrame([[1,5], [2,6], [3,7], [4,-9999]], dtype=float)
        # act
        result = self.test_function(array)
        # assert
        self.assertTrue(result.equals(expected_result))



""" Setup and load the configurations """
import json
import logging
import pathlib
import sys
from dataclasses import dataclass
from typing import List, Optional, Any, Dict, Mapping, Union

import dacite
from marshmallow import Schema, ValidationError, fields, validate

UTILS_CONFIG_LOGGER = logging.getLogger(__name__)

LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
IMPUTATION_STRATEGIES = ["mean", "median",
                         "most_frequent", "constant", "interpolate"]
OPTIMIZE_RANDOMFOREST_MAX_FEATURES = ["sqrt", "log2"]
ALGORITHMS = [
    "randomforest", "xgboost", "singleclass", "us_kmeans",
    "us_kmeans_minibatch", "knn_dtw"
]

PARAMETERS = {
    "app": {
        "algorithm": "randomforest",
        "window": 1024,
        "model": None,
        "model_save": False,  # Save a model file which can be re-used
        "samples": None,
        "log_level": "INFO",
        "threads": -1,
        "imputation": True,
        "imputation_strategy": "mean",
        "imputation_constant": -9999,
        "rasters_are_timeseries": False
    },
    "supervised": {
        "probability": True,
        "all_probabilities": False,
        "remove_outliers": True,  # Remove outliers from the training data
        "optimization": {
            "optimize": False,  # Optimize the model parameters
            "optimize_number": 10,  # Number of iterations for optimization
        },
        "subsample": {
            "active": False,
            "group_by": "class",
            "sample_type": "n",
            "amount": 100,
            "replace": False,
        },
    },
    "unsupervised": {
        "nclasses": 2,  # Number of classes for unsupervised
        "trainfraction": 1.0  # Fraction of raster used for training
    },
    "accuracy": {
        "perform_assesment": True,  # Perform accuracy assessment
        "testfraction": 0.25,  # Fraction of data to use for training
    },
    "dtw": {
        "patterns": None,
        "patterns_save": False,
        "number_of_patterns_per_class": 1,
        "n_neighbors": 1,
        "window": None,
        "max_dist": None,
        "use_pruning": False,
        "penalty": None,
        "optimization_parameters": {
            "n_neighbors": [1],
            "number_of_patterns_per_class": [1],
            "window": [None],
            "penalty": [None]
        }
    },
    "randomforest": {
        "n_estimators": 100,
        "criterion": "gini",
        "max_depth": None,
        "min_samples_split": 2,
        "min_samples_leaf": 1,
        "min_weight_fraction_leaf": 0.,
        "max_features": "sqrt",
        "max_leaf_nodes": None,
        "min_impurity_decrease": 0.,
        "bootstrap": True,
        "oob_score": False,
        "random_state": None,
        "verbose": 0,
        "class_weight": None,
        "ccp_alpha": 0.,
        "max_samples": None,
        "optimization_parameters": {
            "max_features": ['sqrt', 'log2'],
            "max_leaf_nodes": [3, 5, 7],
            "max_depth": [None, 1, 3, 10, 20000],
            "n_estimators": [10, 50, 100],
            "min_samples_split": [2],
            "min_samples_leaf": [1],
            "min_weight_fraction_leaf": [0.0],
            "min_impurity_decrease": [0.0],
            "oob_score": [False],
            "class_weight": [None],
            "ccp_alpha": [0.0],
            "max_samples": [None]
        }
    }
}


class PathField(fields.Field):
    """ Field that serializes to a string and deserializes to a pathlib.Path."""

    def _deserialize(
            self, value: Any, attr: Optional[str],
            data: Optional[Mapping[str, Any]],
            **kwargs: dict) -> pathlib.Path:
        if not isinstance(value, str):
            raise ValidationError(
                "Please provide the path as a string (with quotation marks)")
        return pathlib.Path(value)

    def _serialize(self, value: Any, attr: str, obj: Any, **kwargs: dict)\
            -> str:
        if value is None:
            return ""
        return str(value)


class IntFloatField(fields.Field):
    """ Field that accepts int or float and saves it then
    as either int or float."""

    def __init__(self, val_types: Dict[str, fields.Field],
                 allow_none: bool = False,):
        self.valid_types = val_types
        super().__init__(allow_none=allow_none)

    def _deserialize(self, value: Any, attr: Optional[str],
                     data: Optional[Mapping[str, Any]],
                     **kwargs: dict):
        if isinstance(value, int):
            return self.valid_types['Int'].deserialize(
                value, attr, data, **kwargs)
        if isinstance(value, float):
            return self.valid_types['Float'].deserialize(
                value, attr, data, **kwargs)
        raise ValidationError(
            "Given Value is not Int, Float or None")


def validate_path(path: Any) -> None:
    """Validates if file path exists"""
    path = pathlib.Path(path)
    if not path.exists():
        raise ValidationError('Not a valid path')


def validate_filepath_model(value: Any) -> None:
    """Validates if model path ends on .zip"""
    path = pathlib.Path(value)
    if path.suffix != '.zip':
        raise ValidationError(
            'Model path should include a model.zip at the end')


def validate_filepath_samples(value: Any) -> None:
    """Validates if samples path ends on .csv or .pkl"""
    path = pathlib.Path(value)
    if path.suffix not in ['.csv', '.pkl']:
        raise ValidationError('Not a valid path')


def validate_app_threads(n_threads: int) -> None:
    """Validates number of app threads"""
    if n_threads < -1 or n_threads == 0:
        raise ValidationError(
            "App threads can't be smaller than -1. or equal 0."
            "Choose either -1(flexible number of threads) "
            "or > 0 (fixed number of threads).")


@ dataclass
class AppConfiguration:
    """Dataclass which stores app config"""

    algorithm: str
    # randomforest, xgboost, singleclass,kmeans
    window: int  # Any int, preferably within 2^x
    model: Optional[pathlib.Path]  # Model File location
    model_save: bool  # Save a model file which can be re-used
    samples: Optional[pathlib.Path]  # Samples file location (csv)
    log_level: str  # Logging level
    threads: int  # of threads
    imputation: bool  # Use simple imputation for missing values
    imputation_strategy: str  # Strategy for imputation. mean,
    imputation_constant: int  # constant for imputation
    # if constant was chosen as imputation method
    rasters_are_timeseries: bool  # Rasters are timeseries


class AppConfigurationSchema(Schema):
    """Schema for the App configuration dataclass.
    Will raise ValidationErrors when invalid data are passed in"""

    algorithm = fields.Str(validate=validate.OneOf(ALGORITHMS))
    window = fields.Int(
        validate=validate.Range(min=1, error="Value must be greater than 0"))
    model = PathField(validate=[validate_path, validate_filepath_model],
                      allow_none=True)
    model_save = fields.Bool()
    samples = PathField(
        validate=[validate_path, validate_filepath_samples], allow_none=True)
    threads = fields.Int(validate=validate_app_threads)
    log_level = fields.Str(validate=validate.OneOf(LOG_LEVELS))
    imputation = fields.Bool()
    imputation_strategy = fields.Str(
        validate=validate.OneOf(IMPUTATION_STRATEGIES))
    imputation_constant = fields.Int()
    rasters_are_timeseries = fields.Bool()


@ dataclass
class OptimizeSupervisedConfiguration():
    """Dataclass which stores the optimization parameters for
    the supervised classification"""

    optimize: bool  # Optimize the model parameters
    optimize_number: int  # Number of iterations for optimization


class OptimizeSupervisedConfigurationSchema(Schema):
    """Schema for the optimization part of the supervised classification
    configuration dataclass.
    Will raise ValidationErrors when invalid data are passed in"""

    optimize = fields.Bool()
    optimize_number = fields.Int(
        validate=validate.Range(
            min=1,
            error="Value must be greater than 0. Better value is > 5.")
    )

@ dataclass
class SubsampleSupervisedConfiguration():
    """Dataclass which stores the subsample parameters for
    the supervised classification"""

    active: bool  # Do sampling
    group_by: str  # Groupby class or roi_fid
    sample_type: str # n or frac
    amount: Union[int, float] # int for n and float for frac
    replace: bool # reuse values when sampling


class SubsampleSupervisedConfigurationSchema(Schema):
    """Schema for the subsample part of the supervised classification
    configuration dataclass.
    Will raise ValidationErrors when invalid data are passed in"""
    active = fields.Bool()
    group_by = fields.Str(validate=validate.OneOf(['class', 'roi_fid']))
    sample_type = fields.Str(validate=validate.OneOf(['n', 'frac']))
    amount = IntFloatField(
        {'Int':
            fields.Int(
                validate=validate.Range(
                    min=1,
                    error="Value must be >= 1."),
            ),
         'Float':
            fields.Float(
                validate=validate.Range(
                    min=0.01, max=1))
         }
    )
    replace = fields.Bool()

@ dataclass
class SupervisedConfiguration:
    """Dataclass which stores supervised classification config"""

    probability: bool  # output probability map
    all_probabilities: bool
    remove_outliers: bool  # Remove outliers from the training data
    optimization: OptimizeSupervisedConfiguration
    subsample: SubsampleSupervisedConfiguration


class SupervisedConfigurationSchema(Schema):
    """Schema for the supervised classification configuration dataclass.
    Will raise ValidationErrors when invalid data are passed in"""

    probability = fields.Bool()
    all_probabilities = fields.Bool()
    remove_outliers = fields.Bool()  # Remove outliers from the training data
    optimization = fields.Nested(OptimizeSupervisedConfigurationSchema)
    subsample = fields.Nested(SubsampleSupervisedConfigurationSchema)


@ dataclass
class UnsupervisedConfiguration:
    """Dataclass which stores unsupervised classification config"""

    nclasses: int  # Number of classes for unsupervised
    trainfraction: float  # Fraction of raster used for training


class UnsupervisedConfigurationSchema(Schema):
    """Schema for the unsupervised classification configuration dataclass.
    Will raise ValidationErrors when invalid data are passed in"""

    nclasses = fields.Int(validate=validate.Range(
        min=2, error="Value must be > 1 . This results in at least 2 classes.")
    )  # Number of classes for unsupervised
    trainfraction = fields.Float(validate=validate.Range(
        min=0.01,
        max=1.0,
        error="Fraction of data used for training should be at least 0.01 "
        "and <= 1.0(use all data for training)")
    )  # Fraction of raster used for training


@ dataclass
class OptimizeRandomForestConfiguration:
    """Dataclass which stores algorithm config"""

    n_estimators: List[int]
    max_depth: List[Optional[int]]
    max_features: List[Union[str, int, float]]
    max_leaf_nodes: List[Optional[int]]


class OptimizeRandomForestConfigurationSchema(Schema):
    """Schema for the algorithm configuration dataclass.
    Will raise ValidationErrors when invalid data are passed in"""
    n_estimators = fields.List(
        fields.Int(
            validate=validate.Range(
                min=1,
                error="Value must be >= 1. Rather use 50 or 100.")
        ))
    max_depth = fields.List(
        fields.Int(
            validate=validate.Range(
                min=1,
                error="Value must be >= 1 or NONE."),
            allow_none=True
        ))
    max_features = fields.List(
        fields.Str(
            validate=validate.OneOf(["auto", "sqrt", "log2"]))
    )
    max_leaf_nodes = fields.List(
        fields.Int(
            validate=validate.Range(
                min=1,
                error="Value must be >= 1 or NONE."),
            allow_none=True
        )
    )


@ dataclass
class RandomForestConfiguration:
    """Dataclass which stores algorithm config"""

    n_estimators: int
    criterion: str
    max_depth: Optional[int]
    min_samples_split: Union[int, float]
    min_samples_leaf: Union[int, float]
    min_weight_fraction_leaf: float
    max_features: Union[str, int, float]
    max_leaf_nodes: Optional[int]
    min_impurity_decrease: float
    bootstrap: bool
    oob_score: bool
    random_state: Optional[int]
    verbose: int
    class_weight: Optional[Union[str, dict, List[dict]]]
    ccp_alpha: float
    max_samples: Optional[Union[int, float]]
    optimization_parameters: dict


class RandomForestConfigurationSchema(Schema):
    """Schema for the algorithm configuration dataclass.
    Will raise ValidationErrors when invalid data are passed in"""
    n_estimators = fields.Int(validate=validate.Range(
        min=1, error="Value must be >= 1. Rather use 50 or 100.")
    )
    criterion = fields.Str(validate=validate.OneOf(["gini", "entropy"]))
    max_depth = fields.Int(validate=validate.Range(
        min=1, error="Value must be >= 1 or NONE."), allow_none=True
    )
    min_samples_split = IntFloatField(
        {'Int': fields.Int(
            validate=validate.Range(
                min=1,
                error="Consider min_samples_split as the minimum number")),
         'Float': fields.Float(
             validate=validate.Range(min=0.01, max=1))
         }
    )
    min_samples_leaf = IntFloatField(
        {'Int':
            fields.Int(
                validate=validate.Range(
                    min=1,
                    error="Value must be >= 1."),
            ),
         'Float':
            fields.Float(
                validate=validate.Range(
                    min=0.01, max=1))
         }
    )
    min_weight_fraction_leaf = fields.Float(validate=validate.Range(
        min=0, error="Value must be >= 0")
    )
    max_features = fields.Str(validate=validate.OneOf(["sqrt", "log2"]))
    max_leaf_nodes = fields.Int(validate=validate.Range(
        min=1, error="Value must be >= 1 or NONE."), allow_none=True
    )
    min_impurity_decrease = fields.Float(validate=validate.Range(
        min=0, error="Value must be >=0")
    )
    bootstrap = fields.Bool()
    oob_score = fields.Bool()
    random_state = fields.Int(validate=validate.Range(
        min=1, error="Value must be > 0 or NONE."), allow_none=True
    )
    verbose = fields.Int(validate=validate.Range(
        min=0, error="Value must be >= 0 or NONE."), allow_none=True
    )
    class_weight = fields.Str(validate=validate.OneOf(
        ["balanced", "balanced_subsample"]), allow_none=True)
    ccp_alpha = fields.Float(validate=validate.Range(
        min=0, error="Value must be >=0"), allow_none=True
    )
    max_samples = IntFloatField(
        {'Int':
            fields.Int(
                validate=validate.Range(
                    min=1, error="Value must be >= 1 or NONE.")
            ),
         'Float':
            fields.Float(
                validate=validate.Range(
                    min=0.01, max=1)
            ),
         },
        allow_none=True
    )
    optimization_parameters = fields.Dict(
        keys=fields.Str(),
        values=fields.List(fields.Raw(allow_none=True), allow_none=False),
        allow_none=True)


@ dataclass
class AccuracyConfiguration:
    """Dataclass which stores accuracy config"""

    perform_assesment: bool  # Perform accuracy assessment
    testfraction: float  # Fraction of data to use for training


class AccuracyConfigurationSchema(Schema):
    """Schema for the accuracy configuration dataclass.
    Will raise ValidationErrors when invalid data are passed in"""

    perform_assesment = fields.Bool()  # Perform accuracy assessment
    testfraction = fields.Float(validate=validate.Range(
        min=0,
        max=1.0,
        error="Fraction of data used for testing should be > 0 when "
        "perform_assessment is True and < 1.0")
    )  # Fraction of data to use for training


@ dataclass
class DtwConfiguration:
    """Dataclass which stores dtw config"""

    patterns: Optional[pathlib.Path]
    patterns_save: bool
    number_of_patterns_per_class: int
    n_neighbors: int
    window: Optional[int]
    max_dist: Optional[float]
    use_pruning: Optional[bool]
    penalty: Optional[float]
    optimization_parameters: dict


class DtwConfigurationSchema(Schema):
    """Schema for the dtw configuration dataclass.
    Will raise ValidationErrors when invalid data are passed in"""
    patterns = PathField(
        validate=[validate_path, validate_filepath_samples], allow_none=True)
    patterns_save = fields.Bool()  # Remove outliers from the training data
    number_of_patterns_per_class = fields.Int(
        validate=validate.Range(min=1,
                                error="""Must create at least 1 pattern
                                per class"""))
    n_neighbors = fields.Int(
        validate=validate.Range(min=1,
                                error="Neighbors must be at least >=1"))
    window = fields.Int(
        validate=validate.Range(min=1,
                                error="Window must be at least >=1 or None"),
        allow_none=True
    )
    max_dist = fields.Float(
        validate=validate.Range(min=0.001,
                                error="Max_dist must be at least >0 or None"),
        allow_none=True
    )
    use_pruning = fields.Bool(allow_none=True)
    penalty = fields.Float(
        validate=validate.Range(min=0.001,
                                error="Penalty must be at least >0 or None"),
        allow_none=True
    )
    optimization_parameters = fields.Dict(
        keys=fields.Str(),
        values=fields.List(fields.Raw(allow_none=True)),
        allow_none=True)


class ConfigurationSchema(Schema):
    """ConfigurationSchema for nested json.
    Each section in the json will get validated with a different schema"""

    app = fields.Nested(AppConfigurationSchema)
    supervised = fields.Nested(SupervisedConfigurationSchema)
    unsupervised = fields.Nested(UnsupervisedConfigurationSchema)
    accuracy = fields.Nested(AccuracyConfigurationSchema)
    dtw = fields.Nested(DtwConfigurationSchema)
    randomforest = fields.Nested(RandomForestConfigurationSchema)


@ dataclass
class Configuration:
    """Main Configuration dataclass.
    Has sub dataclasses for the different parts"""
    app: AppConfiguration
    supervised: SupervisedConfiguration
    unsupervised: UnsupervisedConfiguration
    accuracy: AccuracyConfiguration
    dtw: DtwConfiguration
    randomforest: RandomForestConfiguration
    # name and tmp_dir are defined on runtime
    name: str
    tmp_dir: Any


def update_config_dict(config_dict_to_update: dict, updates: dict) -> dict:
    """Updates the nested default config dict with the user
    provided nested config dict

    Args:
        config_dict_to_update (dict): nested default config dict
        updates (dict): nested dict with updates

    Returns:
        dict: updated nested dict
    """
    updated_dict = {}
    for key, value in config_dict_to_update.items():
        if key in updates:
            # key is in the updates
            if isinstance(value, dict):
                # call function recursively with sub dict to find the key
                # to be updated
                updated_dict[key] = update_config_dict(value, updates[key])
            else:
                # update the key
                updated_dict[key] = updates[key]
        else:
            # set the default value for the key, if it is not in the
            # update keys
            updated_dict[key] = config_dict_to_update[key]
    return updated_dict


def setup_config(config_path: Optional[pathlib.Path] = None) -> Configuration:
    """Setup method for the Configuration Dataclass.
    Uses marshmallow schemas to load the json and validate
    types and values. Dacite creates dataclass from that.

    Args:
        config_path (pathlib.Path): path to json config file

    Returns:
        config (Configuration): dataclass containing configs
    """
    if config_path and config_path.is_file():
        # loads json from path
        with open(config_path, encoding='UTF-8') as file:
            config_dict = json.load(file)
            updated_config_dict = update_config_dict(PARAMETERS, config_dict)
            config_json = json.dumps(updated_config_dict)
    else:
        print("No local config file found. Using default config")
        # Use default config dictionary
        config_json = json.dumps(PARAMETERS)
    config = json_to_config(config_json)
    return config


def json_to_config(config_json: str) -> Configuration:
    """Converts the json object to a dataclass with type and value validation

    Args:
        config_json (str): JSON formatted string

    Returns:
        [dataclass]: dataclass containing configs
    """
    # validates the type and range of the config Values
    try:
        result = ConfigurationSchema().load(json.loads(config_json))
    except ValidationError as error:
        UTILS_CONFIG_LOGGER.error(error)
        sys.exit(1)

    # add name and tmp dir
    result['name'] = ''
    result['tmp_dir'] = None

    # creates the Configuration dataclass instance from the checked data
    return dacite.from_dict(data_class=Configuration,
                            data=result
                            # config=dacite.Config(type_hooks=converters),
                            )



from unittest import TestCase
import tempfile

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry import Polygon

from classifier.utils.vector import create_spatial_index, true_intersect


class TestCreateSpatialIndex(TestCase):
    def setUp(self):
        self.test_dir = tempfile.TemporaryDirectory()
        self.test_dir_path = self.test_dir.name

    def tearDown(self):
        # Remove the directory after the test
        self.test_dir.cleanup()
        
    def write_test_polygons(self, polygons):
        # Create GeoDataFrames
        gdf = gpd.GeoDataFrame({'geometry': polygons, 'id': range(len(polygons))})
        self.test_polygons_path = self.test_dir_path + "/test_polygons_all.gpkg"
        gdf.to_file(self.test_polygons_path, driver="GPKG")

    def test_empty_input(self):
        self.write_test_polygons([])
        # Call the function with the empty path
        index = create_spatial_index(self.test_polygons_path)

        # Assert that the index is empty
        self.assertEqual(len(list(index.intersection((-180, -90, 180, 90)))), 0)

    def test_non_empty_input(self):
        # Create a temporary file with sample polygons
        polygons = [
            Polygon([(0, 0), (1, 0), (1, 1), (0, 1)]),
            Polygon([(2, 2), (3, 2), (3, 3), (2, 3)])
        ]
        self.write_test_polygons(polygons)

        # Call the function with the temporary file
        index = create_spatial_index(self.test_polygons_path)

        # Assert that the index contains both polygons
        self.assertEqual(len(list(index.intersection((0, 0, 3, 3)))), 2)


class TestTrueIntersect(TestCase):
    def test_no_overlap(self):
        # Create two polygons with no overlap
        geom_1 = Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
        geom_2 = Polygon([(2, 2), (3, 2), (3, 3), (2, 3)])

        result = true_intersect(geom_1, geom_2)

        self.assertFalse(result)
        
    def test_partial_overlap(self):
        # Create two polygons with partial overlap
        geom_1 = Polygon([(0, 0), (2, 0), (2, 2), (0, 2)])
        geom_2 = Polygon([(1, 1), (3, 1), (3, 3), (1, 3)])

        result = true_intersect(geom_1, geom_2)

        self.assertTrue(result)

    def test_full_overlap(self):
        # Create two polygons with full overlap
        geom_1 = Polygon([(0, 0), (2, 0), (2, 2), (0, 2), (0, 0)])
        geom_2 = Polygon([(1, 1), (3, 1), (3, 3), (1, 3), (1, 1)])

        result = true_intersect(geom_1, geom_2)

        self.assertTrue(result)


"Test runs with different configurations"
import numpy as np
import pytest
import rasterio

from classifier.cli import classification
from classifier.settings import WORKSPACE
from classifier.utils.general import save_config_as_json


@pytest.mark.parametrize('test_arg',
                         [[False,  True, False]])
def test_model_save(runner):
    """
    Test for running a supervised randomforest classification
    on a single raster
    """
    save_config_as_json(threads=1, algorithm="randomforest",
                        remove_outliers=False, model_save=True)

    rois_path = WORKSPACE / 'integration_test_rois.gpkg'
    runner.invoke(classification, [
        "--name", 'test_result', '--overwrite', True,
        '--rois', rois_path, "integration_test_rasters"])
    directory = WORKSPACE / 'test_result' / 'classification.tif'
    with rasterio.open(directory, 'r') as dst:
        result = dst.read(1)
        assert np.all(result[:, 0:4] == 0)
        assert np.all(result[:, 5:10] == 1)


@pytest.mark.parametrize('test_arg',
                         [[False,  True, False]])
def test_perform_accuracy_assessment(runner):
    """
    Test for running a supervised randomforest classification
    on a single raster
    """
    save_config_as_json(threads=1, algorithm="randomforest",
                        remove_outliers=False, perform_assesment=True)

    rois_path = WORKSPACE / 'integration_test_rois.gpkg'
    runner.invoke(classification, [
        "--name", 'test_result', '--overwrite', True,
        '--rois', rois_path, "integration_test_rasters"])
    directory = WORKSPACE / 'test_result' / 'classification.tif'
    with rasterio.open(directory, 'r') as dst:
        result = dst.read(1)
        assert np.all(result[:, 0:4] == 0)
        assert np.all(result[:, 5:10] == 1)


@pytest.mark.parametrize('test_arg',
                         [[False,  True, False]])
def test_perform_optimize(runner):
    """
    Test for running a supervised randomforest classification
    on a single raster
    """
    save_config_as_json(threads=1, algorithm="randomforest",
                        remove_outliers=False, optimize=True)

    rois_path = WORKSPACE / 'integration_test_rois.gpkg'
    runner.invoke(classification, [
        "--name", 'test_result', '--overwrite', True,
        '--rois', rois_path, "integration_test_rasters"])
    directory = WORKSPACE / 'test_result' / 'classification.tif'
    with rasterio.open(directory, 'r') as dst:
        result = dst.read(1)
        assert np.all(result[:, 0:4] == 0)
        assert np.all(result[:, 5:10] == 1)



from unittest import TestCase
import tempfile

import geopandas as gpd
import numpy as np
from numpy.testing import assert_array_equal
import rasterio
from shapely.geometry import Polygon

from classifier.accuracy import get_confidence_map_from_random_forest_model
from classifier.accuracy import collect_classification_and_reference

NOISE_FACTOR = 10

class MockTree:
    def predict(self, data):
        return (data + np.random.rand(*data.shape) * NOISE_FACTOR)[0]

class MockModel:
    estimators_ = [MockTree(), MockTree(), MockTree()]


class ConfidenceMapTestCase(TestCase):
    def setUp(self):
        # arrange
        self.test_function = get_confidence_map_from_random_forest_model
        self.array = np.array([
            [[10, 20, 30], [30, 40, 50]],
            [[50, 60, 70], [70, 80, 90]],
        ])

    def test_confidence_map_ok(self):
        # act
        result = self.test_function(MockModel(), self.array, confidence_level=0.95)
        # assert
        self.assertTrue(all(result.ravel() < 5))
        self.assertEqual(result.shape, self.array.shape[-2:])


class AccuracyAssessment(TestCase):
    def setUp(self):
        self.test_dir = tempfile.TemporaryDirectory()
        self.test_dir_path = self.test_dir.name

        # create the test raster
        self.write_test_raster()

    def tearDown(self):
        # Remove the directory after the test
        self.test_dir.cleanup()

    def write_test_polygons(self, polygons_all, polygons_subset):
        # Create GeoDataFrames
        gdf1 = gpd.GeoDataFrame({'geometry': polygons_all, 'id': range(len(polygons_all))})
        gdf2 = gpd.GeoDataFrame({'geometry': polygons_subset, 'id': range(len(polygons_subset))})

        self.test_polygons_all_path = self.test_dir_path + "/test_polygons_all.gpkg"
        self.test_polygons_subset_path = self.test_dir_path + "/test_polygons_subset.gpkg"
        gdf1.to_file(self.test_polygons_all_path, driver="GPKG")
        gdf2.to_file(self.test_polygons_subset_path, driver="GPKG")

    def write_test_raster(self):
        # Define raster properties
        width = 100
        height = 100
        count = 1
        dtype = np.uint8
        crs = None
        data = np.ones(shape=(height, width), dtype=dtype)

        self.test_raster_path = self.test_dir_path + "fake_raster.tif"
        with rasterio.open(
            self.test_raster_path,
            "w",
            driver="GTiff",
            width=width,
            height=height,
            count=count,
            dtype=dtype,
            crs=crs,
        ) as dst:
            dst.write(data, 1)

    def test_collect_classification_and_reference_all_polygons_overlap(self):
        self.write_test_polygons(
            polygons_all=[
                Polygon([(0, 0), (1, 0), (1, 1), (0, 1)]),
                Polygon([(1, 1), (2, 1), (2, 2), (1, 2)]),
                Polygon([(2, 2), (3, 2), (3, 3), (2, 3)]),
            ],
            polygons_subset=[
                Polygon([(0, 0), (0, 5), (5, 5), (5, 0)]),
                Polygon([(1, 1), (2, 1), (2, 2), (1, 2)]),
                Polygon([(2, 2), (3, 2), (3, 3), (2, 3)]),
            ],
        )

        result_y_pred, result_y_true = collect_classification_and_reference(
            raster=self.test_raster_path,
            rois=self.test_polygons_all_path,
            subset=self.test_polygons_subset_path,
        )

        assert result_y_pred.size == result_y_true.size == 0


    def test_collect_classification_and_reference_some_polygons_overlap(self):
        self.write_test_polygons(
            polygons_all=[
                Polygon([(0, 0), (1, 0), (1, 1), (0, 1)]),
                Polygon([(1, 1), (2, 1), (2, 2), (1, 2)]),
                Polygon([(2, 2), (3, 2), (3, 3), (2, 3)]),
            ],
            polygons_subset=[
                Polygon([(1, 1), (2, 1), (2, 2), (1, 2)]),
                Polygon([(2, 2), (3, 2), (3, 3), (2, 3)]),
            ],
        )

        result_y_pred, result_y_true = collect_classification_and_reference(
            raster=self.test_raster_path,
            rois=self.test_polygons_all_path,
            subset=self.test_polygons_subset_path,
        )

        assert_array_equal(result_y_pred, 1)
        assert_array_equal(result_y_true, 0)


    def test_collect_classification_and_reference_no_polygons_overlap(self):
        self.write_test_polygons(
            polygons_all=[
                Polygon([(0, 0), (1, 0), (1, 1), (0, 1)]),
                Polygon([(1, 1), (2, 1), (2, 2), (1, 2)]),
                Polygon([(2, 2), (3, 2), (3, 3), (2, 3)]),
            ],
            polygons_subset=[
                Polygon([(4, 5), (4, 6), (5, 6), (5, 5)]),
                Polygon([(5, 6), (5, 7), (6, 7), (6, 6)]),
                Polygon([(7, 9), (7, 10), (8, 10), (8, 9)]),
            ],
        )

        result_y_pred, result_y_true = collect_classification_and_reference(
            raster=self.test_raster_path,
            rois=self.test_polygons_all_path,
            subset=self.test_polygons_subset_path,
        )
        assert_array_equal(result_y_pred, np.array([1, 1, 1]))
        assert_array_equal(result_y_true, np.array([0, 1, 2]))



"Training module for classifier"
import logging
import dataclasses
from pathlib import Path
from typing import Any, List, Tuple, Dict

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder

from classifier import __version__ as classifier_version
from classifier.utils.general import dict_product
from classifier.settings import ALGORITHM_DICT
from classifier.samples import Samples
from classifier.utils.config import Configuration
from classifier.utils.general import get_available_model_args

TRAIN_LOGGER = logging.getLogger(__name__)


def train_kmeans(train_array: np.ndarray, config: Configuration) -> dict:
    """Train the kmeans model

        Args:
            train_array (array): A data array with columns (bands) and rows (
            pixels)
            config (Configuration): contains config

        Returns:
            (dict): A dictionary conaining the algorithm name, the trained model
            and an empty label key to match the dicts from the supervised
            classifications
    """
    algorithm = ALGORITHM_DICT[config.app.algorithm]
    n_classes = config.unsupervised.nclasses

    TRAIN_LOGGER.info(
        "Now Training Model. This might take some time...")

    kmeans = algorithm(n_clusters=n_classes)
    kmeans.fit(train_array)
    return {
        'app_algorithm': config.app.algorithm,
        'model': kmeans,
        'labels': None
    }


def train_dataset(samples: Samples, out_dir: Path,
                  config: Configuration) -> Tuple[dict, np.ndarray]:
    """Train the model using a dataset

    Args:
        samples (Samples): Dataset containing features in the columns
            with one column named "class" which contains class labels or numbers
        out_dir (Path): Path where to write dataset file
        config (Configuration): contains config

    Returns:
        model_dict (dict): A dictionary containing the name, model and label
                        encoder.
        test (np.ndarray): A test dataset which was not used during training
    """
    # Split the dataset,
    model, xcols, test, labels = init_model_and_train(
        samples.get_samples(),
        out_dir,
        config,
    )

    model_dict = {'app_algorithm': config.app.algorithm,
                  'model': model,
                  'labels': labels,
                  'names': xcols,
                  'version': classifier_version}
    return model_dict, test


def set_model_parameters(algorithm: str, algorithm_args: dict) -> Any:
    """ Set the model parameters

    Args:
        algorithm (str): internal name of the algorithm (model)
        algorithm_args (dict): Algorithm arguments

    Returns:
        model (Any): parametrized model
    """
    model_type = ALGORITHM_DICT[algorithm]
    model = model_type()
    model_algorithm_args = get_available_model_args(algorithm_args, model_type)
    model.set_params(**model_algorithm_args)
    return model


def get_algorithm_args(
        config: Configuration, dataset: pd.DataFrame, out_dir: Path) -> dict:
    """Fills the algorithm arg dict for the chosen algorithm

    Args:
        config (Configuration): Contains Configuration including algorithm
            name and some parameters
        dataset (pd.Dataframe): Dataset
        out_dir (Path): output directory

    Returns:
        algorithm_args (dict): Contains specific model keyword and arguments
            for the model initialization
    """
    algorithm_name = config.app.algorithm
    algorithm_args = {}

    if algorithm_name == 'randomforest':
        # Random forest
        algorithm_args = dataclasses.asdict(config.randomforest)
    elif algorithm_name == 'xgboost' and len(dataset['class'].unique()) < 3:
        algorithm_args['objective'] = 'binary:logistic'
    elif config.app.rasters_are_timeseries:
        # Get raster count and band count
        raster_count = len(dataset.columns.get_level_values(0).unique())-2
        band_count = len(dataset.columns.get_level_values(1).unique())-1
        fit_params = {'raster_count': raster_count, 'band_count': band_count}

        if algorithm_name == "knn_dtw":
            # KNN
            algorithm_args['fit_params'] = fit_params
            metric_params = {
                'window': config.dtw.window,
                'max_dist': config.dtw.max_dist,
                'use_pruning': config.dtw.use_pruning,
                'penalty': config.dtw.penalty}

            # Dtw function (metric function needs both)
            algorithm_args['metric_params'] = {**fit_params,
                                               **metric_params
                                               }
            algorithm_args['n_neighbors'] = config.dtw.n_neighbors
            algorithm_args['number_of_patterns_per_class'] = \
                config.dtw.number_of_patterns_per_class
            algorithm_args['out_dir'] = out_dir
            algorithm_args['patterns_path'] = config.dtw.patterns
            algorithm_args['patterns_save'] = config.dtw.patterns_save
    algorithm_args['n_jobs'] = config.app.threads
    return algorithm_args


def optimize_model(
        model: Any,
        x_train: pd.DataFrame,
        y_train: pd.DataFrame,
        out_dir: Path,
        config: Configuration) -> Any:
    """_summary_

    Args:
        model (sklearn.model): A sklearn model
        x_train (pd.DataFrame): Train data (in case of dtw also with
            class and roi_fid column)
        y_train (pd.DataFrame): class
        out_dir (Path): output path
        config (Configuration): Configuration

    Returns:
        optimized_model (Sklearn.Model): Model with the optimized
            parameters
    """
    optimize_iters = config.supervised.optimization.optimize_number

    if 'randomforest' in config.app.algorithm:
        optimization_parameters = dataclasses.asdict(
            config.randomforest)['optimization_parameters']
    elif config.app.algorithm == 'knn_dtw':
        # Get the user provided optimization parameters
        # # from the config
        params_from_config = dataclasses.asdict(
            config.dtw)['optimization_parameters']

        # Get the transform parameters for the reshape
        fit_params = model['transformer'].get_params()[
            'fit_params']

        # These two parameters are only influencing the performance
        use_pruning = config.dtw.use_pruning
        max_dist = config.dtw.max_dist

        # The optimization parameter combinations have to
        # be put in a list by hand
        # as sklearn struggles with metric_params unpacking
        optimization_parameters = []

        # Build all combinations of the provided optimization
        # parameters
        for comb in dict_product(params_from_config):
            # Only add combination if there are more or
            # equal patterns to neighbors
            if comb['n_neighbors'] <= comb['number_of_patterns_per_class']:
                # Add to optimization parameters
                optimization_parameters.append(
                    {
                        'n_neighbors': [comb['n_neighbors']],
                        'number_of_patterns_per_class':
                        [comb['number_of_patterns_per_class']],
                        'metric_params': [
                            {**fit_params,
                             'window': comb['window'],
                             'penalty': comb['penalty'],
                             'use_pruning': use_pruning,
                             'max_dist': max_dist
                             }],
                        'fit_params': [fit_params]
                    }
                )
    else:
        # Get the optimization parameters from the model attribute
        optimization_parameters = model.optimization_parameters

    algorithm_args = model.random_optimise(
        optimization_parameters,
        x_train,
        y_train,
        out_dir,
        optimize_iters
    )

    if config.app.algorithm == 'knn_dtw':
        algorithm_args['out_dir'] = out_dir
    if config.dtw.patterns:
        algorithm_args['pattern_path'] = config.dtw.patterns
    if config.dtw.patterns_save:
        algorithm_args['patterns_save'] = config.dtw.patterns_save

    optimized_model = set_model_parameters(
        config.app.algorithm, algorithm_args)

    return optimized_model


def init_model_and_train(
        dataset: pd.DataFrame,
        out_dir: Path,
        config: Configuration
    ) -> Tuple[Any, List[str], pd.DataFrame, Dict[str, str]]:
    """Set the model parameters and train it

    Args:
        dataset (Array) : The dataset for input in the model (array)
        out_dir (Path): The output directory
        config (Configuration): Contains config

    Returns:
        model (Any): Trained sklearn model
        xcols (List[str]): Names of bands
        test (pd.DataFrame): Test dataset
        labels (Dict[str, str]): labels mapped to ascending integers
    """
    optimize = config.supervised.optimization.optimize
    test_size = config.accuracy.testfraction
    algorithm_args = {}

    train, test = train_test_split(dataset, test_size=test_size)
    xcols = [x for x in train.columns
             if 'class' not in x and 'roi_fid' not in x]
    if config.app.algorithm == 'knn_dtw':
        # keep class and roi_fid column for knn_dtw training
        x_train = train
    else:
        x_train = train[xcols]
    y_train = train['class']

    # encode the labels
    label_encoder = LabelEncoder()
    y_train_encoded = label_encoder.fit_transform(y_train)
    labels_encoded = zip(label_encoder.transform(label_encoder.classes_), label_encoder.classes_)
    labels = {str(k): str(v) for k, v in labels_encoded}

    # Get the model/algorithm arguments
    algorithm_args = get_algorithm_args(config, dataset, out_dir)

    # Create the model with the given params
    model = set_model_parameters(config.app.algorithm, algorithm_args)
    # Optimize model
    if optimize:
        model = optimize_model(
            model, x_train, y_train_encoded, out_dir, config)

    # y_train_encoded = sample_labels_encoder.transform(y_train)
    TRAIN_LOGGER.info("Train model ...")
    # Use array repr (x_train.values) to make sure feature names are not used.
    # Except when we use KNN DTW, because that needs the band names.
    if not config.app.algorithm == 'knn_dtw':
        x_train = x_train.values

    model.fit(x_train, y_train_encoded)
    TRAIN_LOGGER.info("Model trained.")
    return model, xcols, test, labels



"Accuracy assessment and plotting for classifier"
import itertools
import logging
import time
from pathlib import Path
from typing import List, Optional, Tuple

import fiona
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from shapely.geometry import shape
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import cohen_kappa_score, confusion_matrix
from sklearn.metrics import classification_report
from tqdm import tqdm

from classifier.utils.raster import clip_raster_on_geometry, ndarray_to_df
from classifier.utils.config import Configuration
from classifier.utils.vector import create_spatial_index, true_intersect

ACCURACY_LOGGER = logging.getLogger(__name__)


def write_confusion_matrix(
        model_dict: dict, test_set: pd.DataFrame, out_dir: Path,
        plot: bool = True, csv: bool = True) -> None:
    """ Uses a sklearn confusion matrix (np 2d array) to write to a csv file
    that also contains overall metrics. Also plots the confusion matrix for
    easier viewing.

        Args:
            model_dict (dict):dictionary with the name, model and label encoder
            test_set (pd.DataFrame): test dataset not used during training
            out_dir (Path): output folder
            plot (bool): Whether or not to plot the figure
            csv (bool): Whether or not to write csv file

        returns:
            nothing
    """
    # Convert dict and dataframe to arrays
    x_test = test_set[[x for x in test_set.columns
                       if 'class' not in x and 'roi_fid' not in x]].values
    y_test = np.ravel(
        test_set[[x for x in test_set.columns if 'class' in x]].values)
    preds = model_dict['model'].predict(x_test)
    labels = {int(k): int(v) for k, v in model_dict['labels'].items()}
    preds_decoded = np.vectorize(lambda x: labels[x])(preds)
    compute_confusion_matrix(y_test, preds_decoded, out_dir, plot, csv)


def compute_confusion_matrix(
        y_test: np.ndarray, preds: np.ndarray, out_dir: Path, plot: bool = True,
        csv: bool = True) -> None:
    """ Uses a sklearn confusion matrix (np 2d array) to write to a csv file
    that also contains overall metrics. Also plots the confusion matrix for
    easier viewing.

        Args:
            y_test (np.array): Ground truth (correct) target values
            preds (np.array): Estimated targets as returned by a classifier
            out_dir (Path): output folder
            plot (bool): Whether or not to plot the figure
            csv (bool):  Whether or not to write the csv
    """

    # Write and plot confusion Matrix
    cm_labels = sorted(list(set(np.unique(y_test)) | set(np.unique(
        preds))))
    kappa = cohen_kappa_score(y_test, preds, labels=cm_labels)
    conf_matrix = confusion_matrix(y_test, preds, labels=cm_labels)

    # Metrics: overall mean accuracy and kappa
    ACCURACY_LOGGER.info("\n####-----Accuracy Assessment-----#####\n")
    metrics_list = []
    metrics_list.append(
        f'Overall Accuracy {np.trace(conf_matrix)/np.nansum(conf_matrix)}'
    )
    metrics_list.append(f'Kappa {kappa:.4f}')
    for metric in metrics_list:
        ACCURACY_LOGGER.info(metric)

    if csv:
        df_cm = pd.DataFrame(
            data=conf_matrix, index=cm_labels, columns=cm_labels)
        # Total of the rows
        df_cm['Total'] = df_cm.sum(axis=1)
        # Accuracy
        df_cm['Accuracy'] = np.diag(df_cm[cm_labels]) / df_cm['Total']
        # Appending the numbers to the df
        total_row = df_cm.sum(axis=0).rename('Total')
        reliability = pd.Series(
            np.diag(conf_matrix) / np.sum(conf_matrix, axis=0),
            index=cm_labels).rename('Reliability')
        df_cm_all = pd.concat(
            [df_cm, pd.DataFrame(total_row).T,
             pd.DataFrame(reliability).T],
            axis=0)
        df_cm_all.to_csv(
            out_dir / 'confusion_matrix.csv', sep=';', float_format='%.4f')

    if plot:
        # Plotting if necessary
        try:
            # Plot normalized confusion matrix
            plt.figure()
            plot_confusion_matrix(conf_matrix,
                                  classes=cm_labels,
                                  normalize=True
                                  )
            plt.tight_layout()
            plt.savefig(out_dir / 'confusion_matrix_plot.png', dpi=150)
        except AttributeError:  # Not all models have FIs, so skip
            pass


def plot_confusion_matrix(conf_matrix: np.ndarray, classes: List[str],
                          normalize: bool = False,
                          cmap: matplotlib.colors.Colormap = plt.cm.Blues) \
        -> None:
    """This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.

    Args:
        conf_matrix (np.ndarray): confusion matrix
        classes (list): Class names
        normalize (bool, optional): Wether to normalize the conf matrix
        cmap (colormap, optional): colormap
    """
    if normalize:
        conf_matrix = conf_matrix.astype('float') /\
            conf_matrix.sum(axis=1)[:, np.newaxis]
    plt.imshow(conf_matrix, interpolation='nearest', cmap=cmap)
    plt.title("Confusion Matrix")
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=90, fontsize=6)
    plt.yticks(tick_marks, classes, fontsize=8)

    fmt = '.2f' if normalize else 'd'
    thresh = conf_matrix.max() / 2.
    for i, j in itertools.product(range(conf_matrix.shape[0]),
                                  range(conf_matrix.shape[1])):
        plt.text(j, i, format(conf_matrix[i, j], fmt),
                 horizontalalignment="center",
                 color="white" if conf_matrix[i, j] > thresh else "black",
                 fontsize=4)

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label')


def plot_feature_importances(model_dict: dict, out_dir: Path) -> None:
    """Plot the feature importances of the forest

    Args:
        model_dict (dict): dict containing the model, metadata and feature names
        out_dir (Path): output folder
    """
    importances = model_dict['model'].feature_importances_
    feat_importances = pd.Series(importances, index=model_dict['names'])

    # only features which have higher importance than 0.01 are plotted
    feat_importances_big = feat_importances[feat_importances > 0.01]

    fig, axis = plt.subplots()
    axis.set_title("Feature importances >0.01 \n(band numbers)")
    axis.set_xlabel("Mean decrease in impurity")

    # Get STD whenever possible
    if model_dict['app_algorithm'] in [
            'randomforest', 'randomforest_ts_metrics']:
        std = pd.Series(
            np.std([tree.feature_importances_ for tree in model_dict[
                'model'].estimators_], axis=0), index=model_dict['names'])
        std = std[feat_importances > 0.01]
        feat_importances_big.sort_values(ascending=False).plot(
            ax=axis, color=['red'], kind='barh', xerr=std)
    else:
        feat_importances_big.sort_values(ascending=False).plot(
            ax=axis, color=['red'], kind='barh')
    fig.tight_layout()
    outfile = out_dir / 'feature_importance.png'
    plt.savefig(outfile, dpi=300)


def collect_classification_and_reference(
        raster: Path,
        rois: Path,
        subset: Path | None = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
    """Collect the classification result and reference class for given rois.
    Ignore rois if it intersects with rois in subset (used for classification). 

    Args:
        raster (Path): raster path
        rois (Path): all rois
        subset (Path | None, optional): rois used for classification. Defaults to None.

    Returns:
        Tuple[np.ndarray, np.ndarray]: class prediction, class reference
    """
    ACCURACY_LOGGER.info("Collect predicted and reference class")
    # Is subset is given, exclude its polygons from accuracy assessment
    if subset:
        rtree_index = create_spatial_index(subset)
    all_y_pred = []
    all_y_true = []
    counter = 0
    with fiona.open(rois, "r") as shapefile:
        for roi in tqdm(shapefile):
            roi_geom = shape(roi['geometry'])
            if subset:
                with fiona.open(subset) as subset_polygons:
                    fids = list(rtree_index.intersection(roi_geom.bounds))
                    # only skip when theres a true intersection
                    if any(
                        true_intersect(
                            roi_geom, shape(subset_polygons[int(fid)]['geometry']))
                        for fid in fids
                    ):
                        counter += 1
                        continue
            roi_samples = clip_raster_on_geometry(raster, roi_geom).flatten()
            if roi_samples is not None and roi_samples.size != 0:
                y_pred = roi_samples.astype(int)
                y_true = int(roi['properties']['id'])

                all_y_pred += list(y_pred)
                all_y_true += len(y_pred) * [y_true]

        all_y_pred = np.array(all_y_pred)
        all_y_true = np.array(all_y_true)

        ACCURACY_LOGGER.info(
            "Using %i groundtruth polygons, excluding %i already used in classifier",
            len(shapefile),
            counter)
    return all_y_pred, all_y_true


def assess_accuracy(
        raster: Path,
        rois: Path,
        out_dir: Path,
        config: Configuration,
        subset: Optional[Path] = None,
    ) -> None:
    """"Accuracy assessment

    Args:
        raster (Path): Path of classication raster
        rois (Path): Path of rois
        out_dir (Path): Path of output
        config (Configuration): Configuration
        subset (Path, optional): Path of subset used for classication
    """
    start = time.time()
    y_pred, y_true = collect_classification_and_reference(
        raster,
        rois,
        subset,
    )

    report = pd.DataFrame(
        classification_report(
            y_true=y_true,
            y_pred=y_pred,
            output_dict=True
            )
    ).T
    report.to_csv(
        out_dir / 'classification_report.csv',
        sep=';',
        float_format='%.4f',
    )

    # write confusion matrix
    compute_confusion_matrix(
        y_test=y_true, preds=y_pred, out_dir=out_dir, plot=True, csv=True)
    config.tmp_dir.cleanup()
    ACCURACY_LOGGER.info(
        "Total run time was %i seconds", (int(time.time() - start)))


def get_confidence_map_from_random_forest_model(
        model: RandomForestRegressor,
        array: np.ndarray,
        confidence_level=0.9,
) -> np.ndarray:
    """Get confidence map from random forest model

    Args:
        model (RandomForestRegressor): the model
        array (np.ndarray): the data array to calculate confidence on.
        confidence_level (float, optional): confidence level [0-1]. \
            Defaults to 0.9.

    Returns:
        np.ndarray: confidence map
    """
    data = ndarray_to_df(array)
    stack = None
    for tree in model.estimators_:
        prediction = tree.predict(data)
        stack = np.vstack(
            [stack, prediction]
        ) if stack is not None else prediction

    quantile_edge = (1 - confidence_level) / 2
    quantile_low = np.quantile(stack, quantile_edge, axis=0)
    quantile_high = np.quantile(stack, 1 - quantile_edge, axis=0)
    return ((quantile_high - quantile_low) / 2).reshape(array.shape[-2:])



"""All the necessary functions for preparing the data for the classification
model can be found here.."""
import logging
import pandas as pd


# Set Logging levels on Rasterio and Fiona to not get too many logging msg
RASTERIO_LOGGER = logging.getLogger('rasterio')
RASTERIO_LOGGER.setLevel(logging.CRITICAL)
FIONA_LOGGER = logging.getLogger('fiona')
FIONA_LOGGER.setLevel(logging.CRITICAL)
DATAPREP_LOGGER = logging.getLogger(__name__)


def check_training_extent() -> None:
    """Check whether the extent of the rois falls outside of the extent of the
        rasters"""
    raise NotImplementedError


def class_counts(samples: pd.DataFrame) -> None:
    """Prints the number of pixels removed after outlier removal

    Args:
        samples (pd.DataFrame): Gathered samples with 'inlier' column

    """
    count_dict = samples.groupby(
        ['class', 'inlier']).size().unstack().to_dict(orient='index')

    # print the header of the table
    to_print = (
        f"\t\t{'Class':<4}"
        f"\t{'#Remaining':<10}"
        f"\t{'Removed':<10}"
        f"{'Remaining fraction':<10}"
    )
    DATAPREP_LOGGER.info("%s", to_print)

    for class_label, pixel_count in list(count_dict.items()):
        remaining = pixel_count[1]
        removed = pixel_count[-1]
        # PRint the row of the table
        to_print = (
            f"\t\t{class_label:<4}"
            f"\t{remaining:<10}"
            f"\t{removed:<10}"
            f"\t{(remaining/(remaining+removed)):<10.2f}"
        )
        DATAPREP_LOGGER.info("%s", to_print)



"""All timeseries functionality can be found here"""
import logging
import warnings
from pathlib import Path
from typing import Optional, Any, List, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from dtaidistance import dtw, dtw_ndim


warnings.simplefilter(action='ignore', category=pd.errors.PerformanceWarning)

TIMESERIES_MODULE_LOGGER = logging.getLogger(__name__)


def dtw_metric(ts1: np.ndarray, ts2: np.ndarray, raster_count: int,
               band_count: int, window: Optional[int] = None,
               max_dist: Optional[float] = None, use_pruning: bool = False,
               penalty: Optional[float] = None) -> Any:
    """Metric function for KNearest-Neighbor classification.
       Uses Dynamic-time-warping (DTW) algorithm as distance function.

        Args:
            ts1 (np.ndarray): timeseries reference
            ts2 (np.ndarray): timeseries sample
            raster_count (int): Number of rasters
            band_count (int): Number of bands
            window (int): window constraint
            max_dist (float): max dist to stop dist computation
            use_pruning (bool): set euclidean dist as max_dist
            penalty (float): penalty to add for compression/expansion

        Returns:
            dtw_dist (float): dtw distance
    """

    if band_count == 1:
        return dtw.distance_fast(
            ts1, ts2,
            window=window,
            max_dist=max_dist,
            use_pruning=use_pruning,
            penalty=penalty
        )
    # nD-case
    ts1 = ts1.reshape(raster_count, band_count)
    ts2 = ts2.reshape(raster_count, band_count)

    # (Dependent) Dynamic Time Warping using multidimensional sequences.
    # For independent DTW (DTW_I) distance, rewrite to
    # use the 1-dimensional version in a loop
    # dtw_ndim.distance(ts1, ts2, use_c=True, max_dist=2.5, window=5)
    # 1d version in loop MUCH SLOWER
    # dtw_i = 0
    # for dim in range(band_count):
    #     dtw_i += dtw.distance(ts1[:, dim], ts2[:, dim])
    return dtw_ndim.distance_fast(
        ts1.astype(np.double), ts2.astype(np.double),
        window=window,
        max_dist=max_dist,
        use_pruning=use_pruning,
        penalty=penalty
    )


def get_characteristic_timeseries_per_class(
        train_dataset: pd.DataFrame,
        patterns_per_class_nr: int) -> List[pd.DataFrame]:
    """ Builds the mean of the band values for each timestep and class.
    If optimize, it uses 5 subsets instead of all the data to build the
    mean values for each of the subsets. Can then be used to find the best
    subset.

    Args:
        train_dataset (pd.DataFrame): _description_
        patterns_per_class_nr (int): _description_

    Returns:
        pd.DataFrame: _description_
    """

    all_mean_values_per_class = []
    train_dataset = train_dataset.sort_index(level=[0, 1])

    for i in range(patterns_per_class_nr):
        # build median per roi
        train_dataset = train_dataset.astype(
            {"roi_fid": int})

        # When multiple patterns per class are created, use
        # subsets of the roi in each run to have variations
        if patterns_per_class_nr > 1:
            mean_per_roi = train_dataset.groupby(
                ['roi_fid']).mean().sample(frac=1/3)
        else:
            mean_per_roi = train_dataset.groupby(
                ['roi_fid']).mean()

        # build median of all same class rois
        mean_per_roi['class'] = mean_per_roi['class'].astype(
            int).astype(str)

        mean_values_per_class = mean_per_roi.groupby(['class']).mean()

        # Dates column becomes index
        mean_values_per_class = mean_values_per_class.stack(0)

        # Set column names to band_names
        # Convert dates to datetime
        mean_values_per_class = mean_values_per_class.reset_index(level=-1)
        mean_values_per_class['Date'] = pd.to_datetime(
            mean_values_per_class['Date'])
        mean_values_per_class = mean_values_per_class.set_index(
            mean_values_per_class['Date'], append=True)
        mean_values_per_class.drop(['Date'], axis=1, inplace=True)

        # extra index "Pattern Set" to
        # distinguish the subset results
        mean_values_per_class["Pattern Set"] = i
        mean_values_per_class = mean_values_per_class.set_index(
            mean_values_per_class['Pattern Set'], append=True)
        mean_values_per_class = mean_values_per_class.drop(
            ['Pattern Set'], axis=1)
        mean_values_per_class = mean_values_per_class.swaplevel(0, 2)

        all_mean_values_per_class.append(mean_values_per_class)
    all_mean_values_per_class = pd.concat(all_mean_values_per_class, axis=0)
    return all_mean_values_per_class


def get_dataset_dates_as_days(
        all_mean_values_per_class: pd.DataFrame) -> Union[
            np.ndarray, List[int]]:
    """Get Dates from dataframe and convert to days, starting at 0.

    Args:
        all_mean_values_per_class (pd.DataFrame):
            Contains a Dataframe with band and dates as column index

    Returns:
        x_days (np.ndarray): contains dates converted to days starting
            at 0.
        date_index (list): contains dates as strings
    """
    date_index = all_mean_values_per_class.index.get_level_values(
        1).unique().tolist()
    # convert to datetime
    dates_datetime = [date.to_pydatetime() for date in date_index]
    # build diff of dates
    date_differences = [
        (dates_datetime[i + 1] - dates_datetime[i]).days
        for i in range(0, len(dates_datetime) - 1)]
    date_differences_added = [0]
    for i, date_diff in enumerate(date_differences):
        date_differences_added.append(
            date_differences_added[i] + date_diff)
    x_days = np.array(date_differences_added)
    return x_days, date_index


def create_training_patterns(
        patterns: pd.DataFrame,
        band_names: List[str],
        class_names: list,
        patterns_per_class: int,
        out_dir: Optional[Path] = None) -> pd.DataFrame:
    """ This function uses values for each class to create training
    patterns. It either uses just the mean values which are already
    given to the function or it uses a GAM (Generalized Adversial Model)
    on the average values for each class. The GAM then creates a smooth
    course of the values (like a linear regression)

    Args:
        patterns (List[pd.DataFrame]): Patterns
        band_names (List[str]): Band Names
        class_names (list): Class Names (here id)
        patterns_per_class (int): Number of patterns per class
        out_dir (Path): Output path

    Returns:
        List[pd.DataFrame]: Contains of the found characteristic timeseries
        found for each class and band
    """
    # Figure for plotting the patterns
    _, axes = plt.subplots(
        len(class_names) * patterns_per_class, len(band_names),
        figsize=(len(band_names) * 15, len(class_names) * 10))
    # pylint: disable=invalid-name

    # Convert the dataset dates to days (starting at 0)
    x_days, date_index = get_dataset_dates_as_days(patterns)

    # Add extra dim for gam gridsearch function
    x_days = np.expand_dims(x_days, axis=1)
    predict_dates = x_days.tolist()

    all_patterns_all_subsets = []

    # Iterate through the multiple run dataframes
    # (multiple patterns per class)
    for subset_pattern_id, subset_pattern_df in patterns.groupby(
            level=0, axis=0):

        subset_pattern_df.index = subset_pattern_df.index.droplevel([0, 1])
        all_patterns_per_subset = []

        class_ids = subset_pattern_df.index.unique()

        # Iterates through alle classes and bands and creates characteristic
        # timeseries for each.
        for class_id_i, class_id in enumerate(class_ids):
            all_patterns_per_class_per_band = []

            # Index for plot
            i = (subset_pattern_id*len(class_ids))+int(class_id_i)

            for j, band_name in enumerate(band_names):
                # Extract values for given class and band
                y_values = subset_pattern_df[band_name][class_id]

                axes[i, j].plot(x_days, y_values, 'r')
                axes[i, j].set_title(
                    f"Class id: {class_id}, Band name: {band_name}")
                axes[i, j].scatter(x_days, y_values,
                                   facecolor='gray', edgecolors='none')

                all_patterns_per_class_per_band.append(y_values)
            values_reshaped = np.array(
                all_patterns_per_class_per_band).flatten('F')

            # Create a dataframe with the results.
            # Columns are the dates and bands
            # 1 Row are the values for 1 class

            # Column names
            columns = []
            for date in date_index:
                for band_name in band_names:
                    columns.append(f"{date} {band_name}")
            # cut it off for only predict dates range
            columns = columns[:len(predict_dates)*len(band_names)]

            # Create Dataframe
            df_values = pd.DataFrame(values_reshaped, columns).T
            df_values['class'] = int(class_id)
            all_patterns_per_subset.append(df_values)

        all_patterns_per_subset = pd.concat(all_patterns_per_subset, axis=0)
        all_patterns_all_subsets.append(all_patterns_per_subset)
    all_patterns_all_subsets = pd.concat(all_patterns_all_subsets, axis=0)
    if out_dir is not None:
        plt.savefig(out_dir / "training_patterns.png", dpi=300)
    return all_patterns_all_subsets


def get_characteristic_timeseries(
        dataset: pd.DataFrame,
        number_of_patterns_per_class: int,
        out_dir: Path) -> pd.DataFrame:
    """ Extract timeseries from every roi and then use timeseries clustering
        to combine them to class characteristic timeseries

        Args:
            dataset (pd.DataFrame): sampled pixels
            number_of_patterns_per_class (int): Number of characteristic
                timeseries per class to create
            out_dir (Path): The output directory

        Returns:
            train (pd.DataFrame):
                columns: dates*bands
                rows: time series values for one class

                Contains the found characteristic time series values
                for each class. This can directly be fed to sklearn.
    """

    class_ids = np.unique(dataset['class'])
    band_names = [column_name for column_name in
                  np.unique(dataset.columns.get_level_values('Band')) if
                  column_name != '']

    # Build the average values for each timestep for each class
    all_patterns_per_class = get_characteristic_timeseries_per_class(
        dataset, number_of_patterns_per_class)

    # Uses the mean values per class to create training patterns per class.
    training_patterns = create_training_patterns(
        all_patterns_per_class,
        band_names,
        class_ids,
        number_of_patterns_per_class,
        out_dir
    )
    return training_patterns
